#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  packetprotocol.py
#  
#  Copyright 2015 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

# pylint: disable=W0613

import threading

import bidict

from twisted.internet.protocol import Factory, ClientFactory, Protocol
from twisted.protocols.basic import Int32StringReceiver

def dprint(msg): # pylint:disable=unused-argument
    pass # Added to stop landscape.io complaining about undefined methods, will be overriden upon import by packetmq.dprint


class PacketProtocol(Int32StringReceiver):
    def __init__(self,parent):
        #super(PacketProtocol,self).__init__()
        self.parent = parent
        self.state = "connMade"
    def connectionMade(self):
        dprint("Connection Made...")
        self.parent.initConnection(self)
        self.parent.factory.addClient(self)
    def connectionLost(self,reason=None): # pylint:disable=unused-argument
        self.setState("disonnected")
        self.parent.lostConnection(self,reason)
        self.parent.factory.delClient(self)
    def stringReceived(self,string):
        dprint("RECV START")
        dprint("Received String...")
        self.parent.recvEncoded(string,self.parent.peerFileno(self))
        dprint("RECV END")
    def sendEncoded(self,data):
        dprint("Sending Encoded over Protocol...")
        self.sendString(bytes(data))
        dprint("Sent Encoded over Protocol")
    def setState(self,state):
        self.state = state
    def getState(self):
        return self.state

class PacketFactory(Factory):
    def __init__(self,parent,proto):
        #super(PacketFactory,self).__init__()
        self.parent = parent
        self.proto = proto
        self.clients = bidict.bidict()
        self.sesscounter = 0
        self.cllock = threading.Lock()
    def startedConnecting(self,arg):
        pass
    def buildProtocol(self,addr): # pylint:disable=unused-argument
        return self.proto(self.parent)
    def addClient(self,client):
        self.cllock.acquire()
        self.clients[self.sesscounter]=client
        self.sesscounter+=1
        self.cllock.release()
    def delClient(self,client):
        self.cllock.acquire()
        if client in self.clients.values():
            del self.clients[:client]
        self.cllock.release()

class ClientPacketFactory(ClientFactory):
    def __init__(self,parent,proto):
        #super(ClientPacketFactory,self).__init__()
        self.parent = parent
        self.proto = proto
        self.clients = bidict.bidict()
        self.sesscounter = 0
        self.cllock = threading.Lock()
    def startedConnecting(self,arg):
        pass
    def buildProtocol(self,addr): # pylint:disable=unused-argument
        return self.proto(self.parent)
    def addClient(self,client):
        self.cllock.acquire()
        self.clients[self.sesscounter]=client
        self.sesscounter+=1
        self.cllock.release()
    def delClient(self,client):
        self.cllock.acquire()
        if client in self.clients.values():
            del self.clients[:client]
        self.cllock.release()

# Memory Protocols

class MemoryPacketProtocol(Protocol):
    def __init__(self,parent):
        #super(MemoryPacketProtocol,self).__init__()
        self.parent = parent
        self.state = "connMade"
    def connectionMade(self):
        self.parent.initConnection(self)
        self.parent.factory.addClient(self)
    def connectionLost(self,reason=None): # pylint:disable=unused-argument
        self.parent.factory.delClient(self)
    def sendEncoded(self,data):
        self.parent.recvEncoded(data)
    def setState(self,state):
        self.state = state
    def getState(self):
        return self.state

class MemoryPacketFactory(Factory):
    def __init__(self,parent,proto):
        #super(MemoryPacketFactory,self).__init__()
        self.parent = parent
        self.proto = proto
        self.clients = bidict.bidict()
        self.sesscounter = 0
        self.cllock = threading.Lock()
    def buildProtocol(self,addr): # pylint:disable=unused-argument
        return self.proto(self.parent)
    def addClient(self,client):
        self.cllock.acquire()
        self.clients[self.sesscounter]=client
        self.sesscounter+=1
        self.cllock.release()
    def delClient(self,client):
        self.cllock.acquire()
        if client in self.clients.values():
            del self.clients[:client]
        self.cllock.release()

class MemoryClientPacketFactory(ClientFactory):
    def __init__(self,parent,proto):
        #super(MemoryPacketFactory,self).__init__()
        self.parent = parent
        self.proto = proto
        self.clients = bidict.bidict()
        self.sesscounter = 0
        self.cllock = threading.Lock()
    def buildProtocol(self,addr): # pylint:disable=unused-argument
        return self.proto(self.parent)
    def addClient(self,client):
        self.cllock.acquire()
        self.clients[self.sesscounter]=client
        self.sesscounter+=1
        self.cllock.release()
    def delClient(self,client):
        self.cllock.acquire()
        if client in self.clients.values():
            del self.clients[:client]
        self.cllock.release()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  packet.py
#  
#  Copyright 2015 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  
try:
    import umsgpack as msgpack
except ImportError:
    try:
        import msgpack
    except ImportError:
        raise ImportError("No msgpack module was found!!")

PROTOCOL_VERSION = 1

def dprint(msg): # pylint:disable=unused-argument
    pass # Added to stop landscape.io complaining about undefined methods, will be overriden upon import by packetmq.dprint

class BasePacket(object):
    def __init__(self):
        self.bypassStateCheck = False
    def dataEncoded(self,arg,*args,**kwargs):
        if isinstance(arg,str):
            return arg
        self.onEncode(arg,*args,**kwargs)
        return msgpack.dumps(arg)
    def dataDecoded(self,arg,*args,**kwargs):
        if isinstance(arg,str):
            return msgpack.loads(arg)
        self.onDecode(arg,*args,**kwargs)
        return arg
    
    def recv(self,data,fromid,to,*args,**kwargs):
        """Called by the Server/Client to process the packet."""
        return self.onReceive(data,fromid,to,*args,**kwargs)
    def send(self,data,to,fromobj,*args,**kwargs):
        """Called by the Server/Client for final modifications before encoding."""
        dprint("1/2 Callbacks called")
        return self.onSend(data,to,fromobj,*args,**kwargs)
    
    # Callbacks
    def onEncode(self,data,to=-1): # pylint:disable=unused-argument
        """Callback to be overriden by subclasses.
        
        Called before every encoding.
        
        :param obj data: Non-Encoded payload"""
        return data
    def onSend(self,data,to,fromobj,*args,**kwargs): # pylint:disable=unused-argument
        """Callback to be overriden by subclasses.
        
        Called before every send.
        
        :param obj data: Non-Encoded payload"""
        dprint("2/2 Callbacks called")
        dprint("Packet sent:")
        dprint("data  : %s"%data)
        dprint("to    : %s"%to)
        dprint("from  : %s"%fromobj)
        #print("*args : %s"%args)
        dprint("kwargs: %s"%kwargs)
        return [data,to,fromobj]
    def onDecode(self,data,fromid=-1): # pylint:disable=unused-argument
        """Callback to be overriden by subclasses.
        
        Called before every decoding."""
        return data
    def onReceive(self,packet,fromid,to,*args,**kwargs): # pylint:disable=unused-argument
        """Callback to be overriden by subclasses.
        
        Called on every received packet of this type.
        
        :param obj packet: decoded packet
        :param int fromid: fileno of the socket from which the packet was received."""
        dprint("Packet received:")
        dprint("packet: %s"%packet)
        dprint("fromid: %s"%fromid)
        dprint("to    : %s"%to)
        #print("*args : %s"%args)
        dprint("kwargs: %s"%kwargs)
        return [packet,fromid,to]

class Packet(BasePacket):
    pass

class EchoPacket(Packet):
    def __init__(self,retType):
        super(EchoPacket,self).__init__()
        self.retType = retType
    def onReceive(self,packet,fromid,to,*args,**kwargs):
        super(EchoPacket,self).onReceive(packet,fromid,to,*args,**kwargs)
        print("Received EchoPacket: %s"%packet)
        to.sendPacket(packet,self.retType,fromid)
        return [packet,fromid,to]

class PrintPacket(Packet):
    def onReceive(self,packet,fromid,to,*args,**kwargs):
        super(PrintPacket,self).onReceive(packet,fromid,to,*args,**kwargs)
        print("Received packet from client #%s: %s"%(to.peerFileno(fromid),packet))
        return [packet,fromid,to] 

class InternalPacket(Packet):
    def __init__(self):
        super(InternalPacket,self).__init__()
        self.bypassStateCheck = True

class HandshakeInitPacket(InternalPacket):
    def onReceive(self,packet,fromid,to,*args,**kwargs):
        super(HandshakeInitPacket,self).onReceive(packet,fromid,to,*args,**kwargs)
        if to.peerObj(fromid).getState() != "connMade":
            print("Connection already made, ignoring")
            return [packet,fromid,to]
        elif packet["version"]!=PROTOCOL_VERSION:
            print("Version mismatch, aborting")
            to.softquit(fromid,"reason.versionmismatch")
        else:
            dprint("Establishing connection (1/3)")
            to.peerObj(fromid).setState("idcheck")
            data = {"adapt":to.registry.adaptivePacketIds,"idmap":dict(to.registry.reg_name_int)}
            to.sendPacket(data,"packetmq:handshake_pubids",fromid)
        return [packet,fromid,to]

class HandshakePubidsPacket(InternalPacket):
    def onReceive(self,packet,fromid,to,*args,**kwargs):
        super(HandshakePubidsPacket,self).onReceive(packet,fromid,to,*args,**kwargs)
        if to.peerObj(fromid).getState() != "idcheck":
            return [packet,fromid,to]
        elif packet["adapt"]==to.registry.adaptivePacketIds:
            to.softquit(fromid,"reason.sameadaptivepolicy")
        elif packet["adapt"] and not to.registry.adaptivePacketIds:
            # Ids have already been sent, no further action is required.
            to.sendPacket(None,"packetmq:handshake_finish",fromid) # finish handshake
            return [packet,fromid,to]
        elif not packet["adapt"] and to.registry.adaptivePacketIds:
            # Ids have been received, insert them into the registry
            if list(to.registry.reg_name_int)!=list(packet["idmap"]):
                to.softquit(fromid,"reason.nonequalpackets")
            to.registry.reg_name_int.update(packet["idmap"])
            to.sendPacket(None,"packetmq:handshake_finish",fromid)
        else:
            to.softquit(fromid,"reason.unknown")
        return [packet,fromid,to]

class HandshakeFinishPacket(InternalPacket):
    def onReceive(self,packet,fromid,to,*args,**kwargs):
        super(HandshakeFinishPacket,self).onReceive(packet,fromid,to,*args,**kwargs)
        to.peerObj(fromid).setState("active")
        to.on_connMade(fromid)
        return [packet,fromid,to]
    def onSend(self,data,to,fromobj,*args,**kwargs):
        super(HandshakeFinishPacket,self).onSend(data,to,fromobj,*args,**kwargs)
        to.setState("active")
        #fromobj.on_connMade(to)
        return [data,to,fromobj]

class SoftquitPacket(InternalPacket):
    def onReceive(self,packet,fromid,to,*args,**kwargs):
        super(SoftquitPacket,self).onReceive(packet,fromid,to,*args,**kwargs)
        to.peerObj(fromid).transport.loseConnection()
        print("Softquit: %s"%packet["reason"])
        return [packet,fromid,to]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  setup.py
#  
#  Copyright 2015 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

from distutils.core import setup

with open('pypi.rst') as file:
    long_description = file.read()

setup(name='packetmq',
      version='0.1.1a1',
      description='Packet based networking',
      long_description=long_description,
      author='notna',
      author_email='notna@apparat.org',
      packages=['packetmq',
                ],
      package_data={'packetmq': ['docs/source/*']},
      requires=["twisted","msgpack","bidict"],
      provides=["packetmq"],
      classifiers=["Development Status :: 4 - Beta",
                   "Framework :: Twisted",
                   "Intended Audience :: Developers",
                   "Intended Audience :: Information Technology",
                   "Intended Audience :: Telecommunications Industry",
                   "License :: OSI Approved",
                   "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
                   "Natural Language :: English",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python",
                   "Programming Language :: Python :: 2",
                   "Programming Language :: Python :: 2.7",
                   "Topic :: Communications",
                   "Topic :: Internet",
                   "Topic :: Software Development",
                   "Topic :: Software Development :: Libraries",
                   "Topic :: Software Development :: Libraries :: Python Modules",
                   ]
     )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- test-case-name: packetmq.test -*-
#
#  __init__.py
#  
#  Copyright 2015 notna <notna@apparat.org>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#  
#  

try:
    #from twisted.internet.endpoints import TCP4ServerEndpoint
    from twisted.internet import reactor
except ImportError:
    pass

import threading
import struct

import bidict

import packetprotocol
import packet

PROTOCOL_VERSION = packet.PROTOCOL_VERSION

DPRINT = False
packet.DPRINT = DPRINT
packetprotocol.DPRINT = DPRINT

def dprint(msg):
    if DPRINT:
        print(msg)
packet.dprint = dprint
packetprotocol.dprint = dprint

TYPE_FMT = "!H" # struct used for encoding the packet type
#TYPE_FMT = "!L" # Uncomment for more packet ids, do not forget to also change MAX_PACKET_ID to 2**32
TYPE_LEN = 2 # Length of the type prefix in bytes

MIN_PACKET_ID = 16 # minimum packet id, used during packet registration; should be smaller than MAX_PACKET_ID but bigger than 0
MAX_PACKET_ID = 2**16 # maximum packet id, used during packet registration; correlates to TYPE_FMT

class PacketRegistry(object):
    def __init__(self,adaptPacketIds=False):
        self.reg_name_int = bidict.bidict()
        self.reg_name_obj = bidict.bidict()
        self.packetId = self.packetInt
        self.adaptivePacketIds = adaptPacketIds
    def addPacket(self,name,obj,numid,bypass_assert=False):
        assert MIN_PACKET_ID<=numid<=MAX_PACKET_ID or bypass_assert
        self.reg_name_int[name]=numid
        self.reg_name_obj[name]=obj
    def delPacket(self,arg):
        del self.reg_name_int[self.packetName(arg)]
        del self.reg_name_obj[self.packetName(arg)]
    def packetStr(self,arg):
        if isinstance(arg,str):
            return arg
        elif isinstance(arg,int):
            return self.reg_name_int[:arg]
        elif isinstance(arg,packet.BasePacket):
            return self.reg_name_obj[:arg]
        else:
            raise TypeError("Invalid type %s"%type(arg))
    packetName = packetStr
    def packetInt(self,arg):
        if isinstance(arg,int):
            return arg
        elif isinstance(arg,str):
            return self.reg_name_int[arg]
        elif isinstance(arg,packet.BasePacket):
            return self.reg_name_int[self.reg_name_obj[:arg]]
        else:
            raise TypeError("Invalid type %s"%type(arg))
    def packetObj(self,arg):
        if isinstance(arg,packet.BasePacket):
            return arg
        elif isinstance(arg,int):
            return self.reg_name_obj[self.reg_name_int[:arg]]
        elif isinstance(arg,str):
            return self.reg_name_obj[arg]
        else:
            raise TypeError("Invalid type %s"%type(arg))
    def registerDefaultPackets(self):
        self.addPacket("packetmq:handshake_init",packet.HandshakeInitPacket(),0,True)
        self.addPacket("packetmq:softquit",packet.SoftquitPacket(),1,True)
        self.addPacket("packetmq:handshake_pubids",packet.HandshakePubidsPacket(),2,True)
        self.addPacket("packetmq:handshake_finish",packet.HandshakeFinishPacket(),3,True)

class Peer(object):
    def __init__(self,registry,proto=packetprotocol.PacketProtocol,factory=packetprotocol.PacketFactory):
        self.registry = registry
        self.factory = factory(self,proto)
        self.skip_handshake = False
    def peerFileno(self,arg):
        if isinstance(arg,int):
            return arg
        elif isinstance(arg,packetprotocol.PacketProtocol) or isinstance(arg,Peer):
            return self.factory.clients[:arg]
        else:
            raise TypeError("Invalid type %s"%type(arg))
    def peerObj(self,arg):
        if isinstance(arg,packetprotocol.PacketProtocol) or isinstance(arg,Peer):
            return arg
        elif isinstance(arg,int):
            return self.factory.clients[arg]
        else:
            raise TypeError("Invalid type %s"%type(arg))
    def initConnection(self,conn):
        if self.skip_handshake:
            self.peerObj(conn).setState("active")
            return
        data = {"version":PROTOCOL_VERSION}
        self.sendPacket(data,"packetmq:handshake_init",conn)
    def lostConnection(self,conn,reason=None):
        pass
    def sendPacket(self,data,dtype,to):
        dprint("SEND START")
        to = self.peerObj(to)
        ptint = self.registry.packetInt(dtype)
        pobj = self.registry.packetObj(dtype)
        #import pdb;pdb.set_trace()
        dprint("Calling send callbacks...")
        data,to,fromobj = pobj.send(data,to,self)
        dprint("dtype: %s"%dtype)
        dprint("Encoding packet...")
        data = pobj.dataEncoded(data,to=to)
        dprint("Packing packet...")
        raw = struct.pack(TYPE_FMT,ptint)+data
        dprint("Start sending...")
        self.sendEncoded(raw,to)
        dprint("SEND END")
    def sendEncoded(self,raw,to):
        dprint("Sending encoded...")
        reactor.callFromThread(self.peerObj(to).sendEncoded,raw)
    def recvPacket(self,data,dtype,fromid):
        #dprint("recvPacket(%s,%s,%s)"%(data,dtype,fromid))
        dprint("recvPacket dtype=%s fromid=%s"%(self.registry.packetStr(dtype),self.peerFileno(fromid)))
        pobj = self.registry.packetObj(dtype)
        if self.peerObj(fromid).getState()=="active" or pobj.bypassStateCheck:
            pobj.recv(data,fromid,self)
        else:
            dprint("Invalid state %s for recv"%self.peerObj(fromid))
    def recvEncoded(self,data,fromid):
        ptraw = data[:TYPE_LEN]
        dprint("RECV encodedpacket:")
        dprint("fromid: %s"%fromid)
        ptint = struct.unpack(TYPE_FMT,ptraw)[0]
        dprint("dtype: %s"%self.registry.packetStr(ptint))
        pobj = self.registry.packetObj(ptint)
        data = data[TYPE_LEN:]
        data = pobj.dataDecoded(data,fromid=fromid)
        dprint("data: %s"%data)
        self.recvPacket(data,pobj,fromid)
    def runAsync(self):
        self.thread = threading.Thread(name="Network reactor thread",target=self.run)
        self.thread.daemon = True
        self.thread.start()
    def run(self):
        reactor.run(installSignalHandlers=0) # Will disable Signalhandlers and thus disable subprocesses, see docs for details
    def stop(self):
        reactor.stop()
    def softquit(self,peer,reason="reason.unknown"):
        peerobj = self.peerObj(peer)
        peerobj.setState("softquitted")
        self.lostConnection(peer,reason)
        self.sendPacket({"reason":reason},"packetmq:softquit",peer)
    def on_connMade(self,conn):
        dprint("Connection Made peer=%s"%self.peerFileno(conn))
    def __del__(self):
        self.stop()

class Server(Peer):
    def __init__(self,registry,proto=packetprotocol.PacketProtocol,factory=packetprotocol.PacketFactory):
        super(Server,self).__init__(registry,proto,factory)
    def listen(self,port):
        reactor.listenTCP(port,self.factory)
    def runAsync(self):
        self.thread = threading.Thread(name="Server thread",target=self.run)
        self.thread.daemon = True
        self.thread.start()

class MemoryServer(Server):
    def __init__(self,registry,proto=packetprotocol.MemoryPacketProtocol,factory=packetprotocol.MemoryPacketFactory):
        super(MemoryServer,self).__init__(registry,proto,factory)
        self.skip_handshake = True
        self.state = "connMade"
    def setState(self,state):
        self.state = state
    def getState(self):
        return self.state
    def listen(self,port):
        raise NotImplementedError("Listen is not allowed on MemoryServers")
    def connectClient(self,client):
        self.factory.addClient(client)
        client.factory.addClient(self)
        self.initConnection(client)
    def disconnectClient(self,client):
        self.factory.delClient(self.peerObj(client))
    def sendPacket(self,data,dtype,to):
        self.peerObj(to).recvPacket(data,dtype,to)
    def sendEncoded(self,raw,to):
        self.peerObj(to).recvEncoded(raw,self)

class Client(Peer):
    def __init__(self,registry,proto=packetprotocol.PacketProtocol,factory=packetprotocol.ClientPacketFactory):
        super(Client,self).__init__(registry,proto,factory)
        self.default_peer = None
    def initConnection(self,conn):
        super(Client,self).initConnection(conn)
        if self.default_peer is None or self.default_peer not in self.factory.clients:
            self.default_peer = conn
    def lostConnection(self,conn,reason=None):
        super(Client,self).lostConnection(conn,reason)
        self.default_peer = None
    def connect(self,address):
        reactor.connectTCP(address[0],address[1],self.factory)
    def runAsync(self):
        self.thread = threading.Thread(name="Client thread",target=self.run)
        self.thread.daemon = True
        self.thread.start()
    def sendPacket(self,data,dtype,to=None):
        if to is None:
            to = self.default_peer
        super(Client,self).sendPacket(data,dtype,to)
    def sendEncoded(self,raw,to=None):
        if to is None:
            to = self.default_peer
        super(Client,self).sendEncoded(raw,to)
    def recvPacket(self,data,dtype,fromid=None):
        if fromid is None:
            fromid = self.default_peer
        super(Client,self).recvPacket(data,dtype,fromid)
    def recvEncoded(self,data,fromid=None):
        if fromid is None:
            fromid = self.default_peer
        super(Client,self).recvEncoded(data,fromid)

class MemoryClient(Client):
    def __init__(self,registry,proto=packetprotocol.MemoryPacketProtocol,factory=packetprotocol.MemoryClientPacketFactory):
        super(MemoryClient,self).__init__(registry,proto,factory)
        self.skip_handshake = True
        self.state = "connMade"
    def setState(self,state):
        self.state = state
    def getState(self):
        return self.state
    def connect(self,server):
        if self.default_peer is None:
            self.default_peer = server
        server.connectClient(self)
        self.initConnection(server)
    def disconnect(self,server=None):
        if server is None:
            server = self.default_peer
        self.lostConnection(server,"reason.generic")
        server.disconnectClient(self)
    def sendPacket(self,data,dtype,to=None):
        if to is None:
            to = self.default_peer
        pobj = self.registry.packetObj(dtype)
        dprint("Calling send callbacks...")
        data,to = pobj.send(data,to,self)
        dprint("dtype: %s"%dtype)
        self.peerObj(to).recvPacket(data,dtype,self)
    def sendEncoded(self,raw,to=None):
        if to is None:
            to = self.default_peer
        self.peerObj(to).recvEncoded(raw,to)
